@@ -8,10 +8,11 @@ use tracing_subscriber::{registry::Registry, prelude::*};
 
 let (chrome_layer, _guard) = ChromeLayerBuilder::new().build();
 tracing_subscriber::registry().with(chrome_layer).init();
-
 ```
 
-!*/
+*/
+
+use crossbeam::channel::Sender;
 
 use tracing::{span, Event, Subscriber};
 use tracing_subscriber::{
@@ -23,9 +24,9 @@ use tracing_subscriber::{
 use json::{number::Number, object::Object, JsonValue};
 use std::{
     marker::PhantomData,
+    path::Path,
     sync::{
         atomic::{AtomicU64, Ordering},
-        mpsc::Sender,
         Arc, Mutex,
     },
 };
@@ -43,6 +44,7 @@ thread_local! {
 
 type NameFn<S> = Box<dyn Fn(&EventOrSpan<'_, '_, S>) -> String + Send + Sync>;
 
+/// A [`Layer`](tracing_subscriber::Layer) that writes a Chrome trace file.
 pub struct ChromeLayer<S>
 where
     S: Subscriber + for<'span> LookupSpan<'span> + Send + Sync,
@@ -58,12 +60,13 @@ where
     _inner: PhantomData<S>,
 }
 
+/// A builder for [`ChromeLayer`](crate::ChromeLayer).
 #[derive(Default)]
 pub struct ChromeLayerBuilder<S>
 where
     S: Subscriber + for<'span> LookupSpan<'span> + Send + Sync,
 {
-    out_file: Option<String>,
+    out_writer: Option<Box<dyn Write + Send>>,
     name_fn: Option<NameFn<S>>,
     cat_fn: Option<NameFn<S>>,
     include_args: bool,
@@ -94,7 +97,7 @@ where
 {
     pub fn new() -> Self {
         ChromeLayerBuilder {
-            out_file: None,
+            out_writer: None,
             name_fn: None,
             cat_fn: None,
             include_args: false,
@@ -104,35 +107,45 @@ where
         }
     }
 
-    /**
-    Set the file to which to output the trace.
+    /// Set the file to which to output the trace.
+    ///
+    /// Defaults to `./trace-{unix epoch time}.json`.
+    ///
+    /// # Panics
+    ///
+    /// If `file` could not be opened/created.
+    pub fn file<P: AsRef<Path>>(self, file: P) -> Self {
+        self.writer(std::fs::File::create(file).expect("Failed to create trace file."))
+    }
 
-    Defaults to "./trace-{unix epoch time}.json"
-    */
-    pub fn file(mut self, file: String) -> Self {
-        self.out_file = Some(file);
+    /// Supply an arbitrary writer to which to write trace contents.
+    ///
+    /// # Examples
+    ///
+    /// ```rust
+    /// # use tracing_chrome::ChromeLayerBuilder;
+    /// # use tracing_subscriber::prelude::*;
+    /// let (layer, guard) = ChromeLayerBuilder::new().writer(std::io::sink()).build();
+    /// # tracing_subscriber::registry().with(layer).init();
+    /// ```
+    pub fn writer<W: Write + Send + 'static>(mut self, writer: W) -> Self {
+        self.out_writer = Some(Box::new(writer));
         self
     }
 
-    /**
-    Include arguments in each trace entry.
-
-    Defaults to false.
-
-    Includes the arguments used when creating a span/event in the "args" section of the trace entry.
-    */
+    /// Include arguments in each trace entry.
+    ///
+    /// Defaults to false. Includes the arguments used when creating a span/event
+    /// in the "args" section of the trace entry.
     pub fn include_args(mut self, include: bool) -> Self {
         self.include_args = include;
         self
     }
 
-    /**
-    Include file+line with each trace entry.
-
-    Defaults to true.
-
-    This can add quite a bit of data to the output so turning it off might be helpful when collecting larger traces.
-    */
+    /// Include file+line with each trace entry.
+    ///
+    /// Defaults to true. This can add quite a bit of data to the output so turning
+    /// it off might be helpful when collecting larger traces.
     pub fn include_locations(mut self, include: bool) -> Self {
         self.include_locations = include;
         self
@@ -186,6 +199,11 @@ where
         self
     }
 
+    /// Creates a [`ChromeLayer`](crate::ChromeLayer) and associated [`FlushGuard`](crate::FlushGuard).
+    ///
+    /// # Panics
+    ///
+    /// If no file or writer was specified and the default trace file could not be opened/created.
     pub fn build(self) -> (ChromeLayer<S>, FlushGuard) {
         ChromeLayer::new(self)
     }
@@ -236,6 +254,7 @@ enum Message {
     Drop,
 }
 
+/// Represents either an [`Event`](tracing::Event) or [`SpanRef`](tracing_subscriber::registry::SpanRef).
 pub enum EventOrSpan<'a, 'b, S>
 where
     S: Subscriber + for<'span> LookupSpan<'span> + Send + Sync,
@@ -249,22 +268,27 @@ where
     S: Subscriber + for<'span> LookupSpan<'span> + Send + Sync,
 {
     fn new(mut builder: ChromeLayerBuilder<S>) -> (ChromeLayer<S>, FlushGuard) {
-        let (tx, rx) = std::sync::mpsc::channel::<Message>();
+        let (tx, rx) = crossbeam::channel::unbounded();
         OUT.with(|val| val.replace(Some(tx.clone())));
 
-        let out_file = builder.out_file.unwrap_or(format!(
-            "./trace-{}.json",
-            std::time::SystemTime::UNIX_EPOCH
-                .elapsed()
-                .unwrap()
-                .as_secs()
-        ));
+        let out_writer = builder.out_writer.unwrap_or_else(|| {
+            Box::new(
+                std::fs::File::create(format!(
+                    "./trace-{}.json",
+                    std::time::SystemTime::UNIX_EPOCH
+                        .elapsed()
+                        .unwrap()
+                        .as_secs()
+                ))
+                .expect("Failed to create trace file."),
+            )
+        });
 
         let handle = std::thread::spawn(move || {
-            let write = std::fs::File::create(out_file).unwrap();
-            let mut write = BufWriter::new(write);
-            write.write_all(b"[").unwrap();
+            let mut write = BufWriter::new(out_writer);
+            write.write_all(b"[\n").unwrap();
 
+            let mut has_started = false;
             for msg in rx {
                 if let Message::Flush = &msg {
                     write.flush().unwrap();
@@ -330,9 +354,15 @@ where
                     }
                 }
 
+                if has_started {
+                    write.write_all(b",\n").unwrap();
+                }
                 write.write_all(entry.dump().as_bytes()).unwrap();
-                write.write_all(b",\n").unwrap();
+                has_started = true;
             }
+
+            write.write_all(b"\n]").unwrap();
+            write.flush().unwrap();
         });
 
         let guard = FlushGuard {
@@ -418,7 +448,8 @@ where
     }
 
     fn get_root_id(span: SpanRef<S>) -> u64 {
-        span.from_root()
+        span.scope()
+            .from_root()
             .take(1)
             .next()
             .unwrap_or(span)
@@ -466,14 +497,28 @@ where
     S: Subscriber + for<'span> LookupSpan<'span> + Send + Sync,
 {
     fn on_enter(&self, id: &span::Id, ctx: Context<'_, S>) {
-        let ts = self.get_ts();
         if let TraceStyle::Async = self.trace_style {
             return;
         }
 
+        let ts = self.get_ts();
         self.enter_span(ctx.span(id).expect("Span not found."), ts);
     }
 
+    fn on_record(&self, id: &span::Id, values: &span::Record<'_>, ctx: Context<'_, S>) {
+        if self.include_args {
+            let span = ctx.span(id).unwrap();
+            let mut exts = span.extensions_mut();
+
+            let args = exts.get_mut::<ArgsWrapper>();
+
+            if let Some(args) = args {
+                let args = Arc::make_mut(&mut args.args);
+                values.record(&mut JsonVisitor { object: args });
+            }
+        }
+    }
+
     fn on_event(&self, event: &Event<'_>, _ctx: Context<'_, S>) {
         let ts = self.get_ts();
         let callsite = self.get_callsite(EventOrSpan::Event(event));
@@ -481,15 +526,14 @@ where
     }
 
     fn on_exit(&self, id: &span::Id, ctx: Context<'_, S>) {
-        let ts = self.get_ts();
         if let TraceStyle::Async = self.trace_style {
             return;
         }
+        let ts = self.get_ts();
         self.exit_span(ctx.span(id).expect("Span not found."), ts);
     }
 
-    fn new_span(&self, attrs: &span::Attributes<'_>, id: &span::Id, ctx: Context<'_, S>) {
-        let ts = self.get_ts();
+    fn on_new_span(&self, attrs: &span::Attributes<'_>, id: &span::Id, ctx: Context<'_, S>) {
         if self.include_args {
             let mut args = Object::new();
             attrs.record(&mut JsonVisitor { object: &mut args });
@@ -501,15 +545,16 @@ where
             return;
         }
 
+        let ts = self.get_ts();
         self.enter_span(ctx.span(id).expect("Span not found."), ts);
     }
 
     fn on_close(&self, id: span::Id, ctx: Context<'_, S>) {
-        let ts = self.get_ts();
         if let TraceStyle::Threaded = self.trace_style {
             return;
         }
 
+        let ts = self.get_ts();
         self.exit_span(ctx.span(&id).expect("Span not found."), ts);
     }
 }
